#include <mpi.h>
#include <memory>
#include <algorithm>
#include <fstream>

struct Communicator {
    enum {
        Previous = -3,
        Next = -2,
        Any = -1,
        Master = 0
    };

    struct DataPack {
        DataPack( size_t size ) :
            data( new char[ size ] ),
            size( size )
        {}

        void copyFrom( const char *source, size_t s ) {
            std::copy( source, source + std::min( size, s ), data.get() );
        }

        void copyTo( char *destination, size_t s ) {
            std::copy( data.get(), data.get() + std::min( size, s ), destination );
        }

        MPI_Request request;
        std::unique_ptr< char[] > data;
        size_t size;
    };

    struct Request {
        Request( DataPack *pack = nullptr ) :
            _pack( pack ),
            _finished( false )
        {}
        Request( Request &&other ) :
            _pack( std::move( other._pack ) ),
            _finished( other._finished )
        {}

        Request &operator=( Request &&other ) {
            using std::swap;
            swap( _pack, other._pack );
            _finished = other._finished;
        }

        bool finished() const {
            return _finished;
        }

        void abort() {
            destroyResources();
        }

        explicit operator bool() const {
            return finished();
        }

        bool test() {
            int result;
            if ( !finished() && valid() ) {
                MPI_Test( &pack().request, &result, MPI_STATUS_IGNORE );
                _finished = result != 0;
            }
            return finished();
        }

        void wait() {
            if ( !finished() && valid() ) {
                MPI_Wait( &pack().request, MPI_STATUS_IGNORE );
                _finished = true;
            }
        }
        bool valid() const {
            return bool( _pack );
        }
    protected:
        void destroyResources() {
            _pack = nullptr;
        }

        DataPack &pack() {
            return *_pack;
        }
    private:
        std::unique_ptr< DataPack > _pack;
        bool _finished;
    };

    struct SendRequest : Request {
        SendRequest( DataPack *pack = nullptr ) :
            Request( pack )
        {}
        SendRequest( SendRequest && ) = default;
        ~SendRequest() {
            wait();
        }

        SendRequest &operator=( SendRequest && ) = default;

        bool test() {
            if ( Request::test() ) {
                destroyResources();
                return true;
            }
            return false;
        }
        void wait() {
            Request::wait();
            destroyResources();
        }
    };

    struct ReceiveRequest : Request {
        ReceiveRequest( DataPack *pack = nullptr ) :
            Request( pack )
        {}
        ReceiveRequest( ReceiveRequest && ) = default;
        ~ReceiveRequest() {
            wait();
        }

        ReceiveRequest &operator=( ReceiveRequest && ) = default;

        template< typename T >
        void fetch( T &destination ) {
            if ( valid() ) {
                if ( !finished() )
                    wait();
                pack().copyTo( reinterpret_cast< char * >( &destination ), sizeof( T ) );
            }
        }

        template< typename T >
        T fetch() {
            T value;
            fetch( value );
            return value;
        }
    };

    Communicator( int *argc, char ***argv ) {
        MPI_Init( argc, argv );
        MPI_Comm_size( MPI_COMM_WORLD, &_worldSize );
        MPI_Comm_rank( MPI_COMM_WORLD, &_id );
    }

    ~Communicator() {
        MPI_Finalize();
    }

    bool master() const {
        return _id == 0;
    }

    int id() const {
        return _id;
    }

    int size() const {
        return _worldSize;
    }

    int address( int en ) {
        if ( en >= 0 && en < size() )
            return en;
        switch( en ) {
        case Previous:
            return id() == 0 ? size() - 1 : id() - 1;
        case Next:
            return id() + 1 == size() ? 0 : id() + 1;
        case Any:
        default:
            return MPI_ANY_SOURCE;
        }
    }

    template< typename T >
    SendRequest send( const T &value, int destination, int tag = Any ) {
        const char *data = reinterpret_cast< const char * >( &value );
        size_t size = sizeof( T );
        DataPack *pack = new DataPack( size );
        pack->copyFrom( data, size );
        MPI_Isend(
            pack->data.get(),
            size,
            MPI_BYTE,
            address( destination ),
            tag == Any ? 0 : tag,
            MPI_COMM_WORLD,
            &pack->request );
        return SendRequest( pack );
    }

    template< typename T >
    ReceiveRequest receive( int source = Any, int tag = Any ) {
        size_t size = sizeof( T );
        DataPack *pack = new DataPack( size );
        MPI_Irecv(
            pack->data.get(),
            size,
            MPI_BYTE,
            address( source ),
            tag == Any ? MPI_ANY_TAG : tag,
            MPI_COMM_WORLD,
            &pack->request );
        return ReceiveRequest( pack );
    }

private:

    int _id;
    int _worldSize;
};
