#include <iostream>
#include <algorithm>
#include <unistd.h>
#include <stdexcept>
#include <queue>
#include <sstream>

#include "communicator.h"

struct Worker {

    enum class Tag {
        Data = 8
    };

    Worker( Communicator &comm, double K, int nextGeneration = 2 ) :
        _comm( comm ),
        _K( K ),
        _nextGeneration( nextGeneration ),
        _generator( std::random_device()() ),
        _distReal( 0, 1 ),
        _distInt( 0, comm.size() - 1 )
    {
        if ( _K >= 1 || _K <= 0 )
            throw std::invalid_argument( "K has to be in range (0; 1)." );
        if ( _nextGeneration <= 0 )
            throw std::invalid_argument( "nextGeneration has to be higher that 0" );

        _recReq = _comm.receive< double >( Communicator::Any, int( Tag::Data ) );

        _queue.push( K );
    }

    ~Worker() {
        for ( auto & r : _requests )
            r.abort();
        _recReq.abort();
    }

    bool finished() {
        /// TO BE IMPLEMENTED
    }

    void next( double value ) {
        for ( int i = 0; i < _nextGeneration; ++i ) {
            if ( random() < value ) {
                std::cout << _comm.id() << " has generated child" << std::endl;
                _requests.emplace_back(
                    _comm.send( value * _K, randomRecipient(), int( Tag::Data ) )
                );
            }
        }
    }

    void run() {
        while ( !finished() ) {
            receive();

            while ( !_queue.empty() ) {
                next( _queue.front() );
                _queue.pop();
            }
            clearRequests();
        }
    }

    void receive() {
        while ( _recReq.test() ) {
            _queue.push( _recReq.fetch< double >() );
            _recReq = _comm.receive< double >( Communicator::Any, int( Tag::Data ) );
        }
    }

    void clearRequests() {
        for ( auto i = _requests.begin(); i != _requests.end(); ) {
            if ( i->test() ) // object is no longer needed
                i = _requests.erase( i );
            else
                ++i;
        }
    }

    double random() {
        return _distReal( _generator );
    }

    int randomRecipient() {
        return _distInt( _generator );
    }
private:
    Communicator &_comm;
    double _K;
    int _nextGeneration;
    std::mt19937 _generator;
    std::uniform_real_distribution<> _distReal;
    std::uniform_int_distribution<> _distInt;
    std::queue< double > _queue;
    std::vector< Communicator::SendRequest > _requests;
    Communicator::ReceiveRequest _recReq;
};

template< typename T >
bool readArg( const char *arg, T &value ) {
    std::istringstream s( arg );
    s >> value;
    return s;
}

int main( int argc, char **argv ) {
    Communicator comm( &argc, &argv );

    double K;
    int nextGeneration;

    if ( argc <= 1 || !readArg( argv[ 1 ], K ) )
        K = 0.5;
    if ( argc <= 2 || !readArg( argv[ 2 ], nextGeneration ) )
        nextGeneration = 2;

    try {

        Worker( comm, K, nextGeneration ).run();
    } catch( std::exception &ex ) {
        std::cerr << "Exception: " << ex.what() << std::endl;
    }
    return 0;
} 